=begin
(c) TIG 2010
Permission to use, copy, modify, and distribute this software for 
any purpose and without fee is hereby granted, provided that the above
copyright notice appear in all copies.
THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

  FreeRotate.rb >> Plugins Folder

Usage:
Make a Selection.
Pick Menu > Plugins > Free Rotate.
  or type freerotate into Ruby Console.
The Selection can consist of any entities - it is temporarily Grouped.
The Center of the Selection's Bounds are used as the 'Pivot', unless there 
is a Guide-Point [@cpoint] in the Selection, then that is used instead.
NB: it uses the first @cpoint found if there are more than one in the Selection, 
so if you want to use this tool with several @cpoints then make a group of them 
first - so then only the group accessed].
You move the Cursor and the Selection Rotates in 3D about the Pivot, 
following the Cursor... 
Rotation is entirely 'free' but it can be 'constrained' if you let the Cursor 
infer-snap to edges, faces or one of the three main 'planes' [RGB/XYZ]...

Left-Click = Confirm Rotation.
Esc         = Cancel.
Select another Tool = Cancel.

The Selection remains after the tool ends.
The operation will Undo in one step.
For your assistance a temporary cline Selection bounding-box is added 
[Note: this can be snapped to].
A rubber-banding cline is also shown from the Pivot to the Cursor as it moves.

Donations:
by PayPal to info @ revitrev.org

Version:
1.1 20100825 First release.
=end
#-----------------------------------------------------------------------------
require 'sketchup.rb'
#-----------------------------------------------------------------------------
class FreeRotate

def getExtents
    bbox=Sketchup.active_model.bounds
    bbox.add(@ip.position)if @ip and @ip.valid?
    bbox.add(@ip0.position)if @ip0 and @ip0.valid?
    bbox.add(@ip1.position)if @ip1 and @ip1.valid?
    return bbox
end

# This is the standard Ruby initialize method that is called when you create
# a new object.
def initialize
    @ip=nil
    @ip0=nil
    @ip1=nil
end

# The activate method is called by SketchUp when the tool is first selected.
# it is a good place to put most of your initialization
def activate
    # The Sketchup::InputPoint class is used to get 3D points from screen
    # positions.  It uses the SketchUp inferencing code.
    # In this tool, we will have two points for the endpoints of the line.
    @ip=Sketchup::InputPoint.new
    @ip0=nil
    @ip1=Sketchup::InputPoint.new
    @tline=nil
    @group=nil
    @pivot=ORIGIN
    ###############################
    @model=Sketchup.active_model
    @ents=@model.active_entities
    @sel=@model.selection
    if not @sel[0]
      self.onCancel(nil,nil)
      return nil
    end#if
    @sela=@sel.to_a
    Sketchup.active_model.start_operation("Free Rotate")
    @sel.clear
    @cpoint=nil
    @sela.each{|e|
      if e.class==Sketchup::ConstructionPoint
        @cpoint=e
        @sela=@sela-[e]
        break
      end#if
    }
    ###
    @group=@ents.add_group(@sela)
    ###bb=Geom::BoundingBox.new
    bb=@group.bounds
    if @cpoint
      @pivot=@cpoint.position
    else
      @pivot=bb.center
    end#if
    ### make temporary cline box
    @tbox=[]
    @tbox << @ents.add_cline(bb.corner(0),bb.corner(1))
    @tbox << @ents.add_cline(bb.corner(0),bb.corner(2))
    @tbox << @ents.add_cline(bb.corner(0),bb.corner(4))
    @tbox << @ents.add_cline(bb.corner(1),bb.corner(3))
    @tbox << @ents.add_cline(bb.corner(1),bb.corner(5))
    @tbox << @ents.add_cline(bb.corner(2),bb.corner(3))
    @tbox << @ents.add_cline(bb.corner(2),bb.corner(6))
    @tbox << @ents.add_cline(bb.corner(3),bb.corner(7))
    @tbox << @ents.add_cline(bb.corner(4),bb.corner(5))
    @tbox << @ents.add_cline(bb.corner(4),bb.corner(6))
    @tbox << @ents.add_cline(bb.corner(7),bb.corner(5))
    @tbox << @ents.add_cline(bb.corner(7),bb.corner(6))
    ###
    @msg="Free Rotate: With Cursor, Left-Click=Accept, Esc=Cancel, Alt=Pick-Pivot-Point..."
    Sketchup::set_status_text(@msg)
    self.reset(nil)
end

# deactivate is called when the tool is deactivated because
# a different tool was selected etc
def deactivate(view)
    self.undo_rotation()if @state!=1
    self.reselect()
    Sketchup::set_status_text("",SB_PROMPT)
    Sketchup::set_status_text("",SB_VCB_LABEL)
    Sketchup::set_status_text("",SB_VCB_VALUE)
    view.invalidate if view
    return nil
end

# onCancel is called when the user hits the escape key
def onCancel(flag, view)
    self.undo_rotation()if view
    Sketchup::set_status_text("",SB_PROMPT)
    Sketchup::set_status_text("",SB_VCB_LABEL)
    Sketchup::set_status_text("",SB_VCB_VALUE)
    view.invalidate if view
    Sketchup.send_action("selectSelectionTool:")
    return nil
end

# The onMouseMove method is called whenever the user moves the mouse.
# because it is called so often, it is important to try to make it efficient.
# In a lot of tools, your main interaction will occur in this method.
def onMouseMove(flags, x, y, view)
    if not @ip0 ### base point
        @ip0=Sketchup::InputPoint.new
        @ip.pick(view, x, y)
        if @ip != @ip0
          view.invalidate if @ip.display? or @ip0.display?
          @ip0.copy!(@ip)
        end
    end#if
    view.tooltip="Free-Rotate"#+@ip1.tooltip
    if @state != 1
        Sketchup::set_status_text(@msg)if @state==0
        @ip.pick(view, x, y)
        if @ip != @ip1
          view.invalidate if @ip.display? or @ip1.display?
          @ip1.copy!(@ip)
          self.temp_rotation()if @state==0
        end
    end
end

# The onLButtonUp method is called when the user releases the left mouse button.
def onLButtonDown(flags, x, y, view)
    @tline.erase! if @tline and @tline.valid?
    self.commit_rotation() if @state==0
end

# onKeyDown is called when the user presses a key on the keyboard.
# We are checking it here to see if the user pressed the shift key
# so that we can do inference locking
def onKeyDown(key, repeat, flags, view)
    if( key == CONSTRAIN_MODIFIER_KEY && repeat == 1 )
        @shift_down_time=Time.now
        # if we already have an inference lock, then unlock it
        if( view.inference_locked? )
            # calling lock_inference with no arguments actually unlocks
            view.lock_inference
        elsif( @state == 0 && @ip1.valid? )
            view.lock_inference @ip1
        elsif( @state == 1 && @ip2.valid? )
            view.lock_inference @ip2, @ip1
        end
    end
end

# onKeyUp is called when the user releases the key
# We use this to unlock the interence
# If the user holds down the shift key for more than 1/2 second, then we
# unlock the inference on the release.  Otherwise, the user presses shift
# once to lock and a second time to unlock.
def onKeyUp(key, repeat, flags, view)
    if( key == CONSTRAIN_MODIFIER_KEY &&
        view.inference_locked? &&
        (Time.now - @shift_down_time) > 0.5 )
        view.lock_inference
    end
end

def enableVCB?
   return false
end

def resume(view=nil)
    view.invalidate if view
    if @state==0
      Sketchup::set_status_text(@msg)
    end#if
end

# The draw method is called whenever the view is refreshed.  
# It lets the tool draw any temporary geometry that it needs to.
def draw(view)
    @tline.erase! if @tline and @tline.valid?
    if @state!=1 and (@ip.valid? or @ip1.valid?)
        @ip1.draw(view) if @ip1.display?
        view.set_color_from_line(@pivot, @ip1.position)
        view.draw_line(@pivot, @ip1.position)
        @tline=@ents.add_cline(@pivot, @ip1.position)
    end
end

###############################################

# Reset the tool back to its initial state
def reset(view=nil)
    # This variable keeps track of which point we are currently getting
    @state=0
    @ip.clear
    @ip1.clear
    if view
        view.tooltip=nil
        view.lock_inference
        view.invalidate
    end
end

def commit_rotation()
    @state=1
    @tline.erase! if @tline and @tline.valid?
    @tbox.each{|e|e.erase! if e.valid?}if @tbox
    @group.explode if @group and @group.valid?
    self.reselect()
    Sketchup.active_model.commit_operation
    Sketchup.send_action("selectSelectionTool:")
    return nil
end

def temp_rotation()
  return nil if @pivot==@ip0.position
  return nil if @pivot==@ip1.position
  ###
  p0=@ip0.position
  p1=@ip1.position
  if @pivot != p0 and @pivot != p1 and p0 != p1
    vec_p_0=@pivot.vector_to(p0)
    vec_p_1=@pivot.vector_to(p1)
    axis=vec_p_0.cross(vec_p_1)
    return nil if axis.length==0
    ang=vec_p_0.angle_between(vec_p_1)
    return nil if ang==0
    tr=Geom::Transformation.rotation(@pivot,axis,ang)
    @group.transform!(tr)
    @ip0.copy!(@ip1)
  end
    ###
end

def undo_rotation()
    @state=1
    @tline.erase! if @tline and @tline.valid?
    @tbox.each{|e|e.erase! if e.valid?}if @tbox
    @group.explode if @group and @group.valid?
    Sketchup.active_model.commit_operation
    Sketchup.send_action("editUndo:")###########
    self.reselect()
    Sketchup.send_action("selectSelectionTool:")
    return nil
end

def reselect()
    @sela.each{|e|@sel.add(e)if e.valid?}if @sela
    @sel.add(@cpoint)if @cpoint
end

end # class

#-----------------------------------------------------------------------------
# shortcut
def freerotate()
    Sketchup.active_model.select_tool FreeRotate.new
end
######-----------------------------------------------

### add menu item etc #####################################################
if not file_loaded?(File.basename(__FILE__))
  ### make command1
  text10 = "Free Rotate"
  text11 = "Free Rotate"
  text12 = "Free Rotate Selection about a Point with Cursor"+"..."
  cmd1=UI::Command.new(text10){freerotate()}
  cmd1.tooltip=text11
  cmd1.status_bar_text=text12
  UI.menu("Plugins").add_item(cmd1)
  ###
end#if
file_loaded(File.basename(__FILE__))
###
